from typing import Optional


class AlreadyQueuedError(Exception):
    """
    Raised when a task has already been enqueued and may not be added to the
    queue.

    The exception may have the property `likely_culprit` set. If it is, this
    is the Celery Task ID of the task _most likely_ holding onto the lock.

    `likely_culprit` is here to assist in debugging deadlocks. Retrieving this
    value is not atomic, and thus should not be relied upon.
    """
    def __init__(self, *, expires_in: Optional[int] = None,
                 likely_culprit: Optional[str] = None):
        super().__init__()
        self.likely_culprit = likely_culprit
        self.expires_in = expires_in

    def __repr__(self):
        return (
            '<AlreadyQueuedError('
            f'likely_culprit={self.likely_culprit!r},'
            f' expires_in={self.expires_in!r})>'
        )



import enum

from sqlalchemy import (
    Column,
    TIMESTAMP,
    Integer,
    String,
    func,
    BigInteger,
    text,
    Enum
)
from sqlalchemy.orm import declarative_base, sessionmaker

Base = declarative_base()
Session = sessionmaker()


class WorkerStatus(enum.Enum):
    #: Worker is answering heartbeats.
    ALIVE = 0
    #: We didn't get an offline event, but we're not getting heartbeats.
    LOST = 10
    #: We got a shutdown event for the worker.
    OFFLINE = 20


class TaskStatus(enum.Enum):
    RECEIVED = 0
    STARTED = 10
    SUCCEEDED = 20
    FAILED = 30
    REJECTED = 40
    REVOKED = 50
    RETRIED = 60


class TaskInstance(Base):
    __tablename__ = 'task_instance'

    uuid = Column(String, primary_key=True)
    name = Column(String)
    status = Column(Enum(TaskStatus))
    hostname = Column(String)
    args = Column(String)
    kwargs = Column(String)

    runtime = Column(Integer)

    received = Column(TIMESTAMP, server_default=func.now())
    started = Column(TIMESTAMP, nullable=True)
    failed = Column(TIMESTAMP, nullable=True)
    rejected = Column(TIMESTAMP, nullable=True)
    succeeded = Column(TIMESTAMP, nullable=True)

    retries = Column(Integer, server_default=text('0'))
    last_seen = Column(TIMESTAMP, server_onupdate=func.now())


class Worker(Base):
    __tablename__ = 'worker'

    #: The hostname of a worker is used as its ID.
    id = Column(String, primary_key=True)

    #: How often the worker is configured to send heartbeats.
    frequency = Column(Integer, server_default=text('0'))
    #: Name of the worker software
    sw_identity = Column(String, nullable=True)
    #: Version of the worker software.
    sw_version = Column(String, nullable=True)
    #: Host operating system of the worker.
    sw_system = Column(String, nullable=True)

    #: Number of currently executing tasks.
    active = Column(BigInteger, server_default=text('0'))
    #: Number of processed tasks.
    processed = Column(BigInteger, server_default=text('0'))

    #: Last known status of the worker.
    status = Column(Enum(WorkerStatus))

    first_seen = Column(TIMESTAMP, server_default=func.now())
    last_seen = Column(TIMESTAMP, server_onupdate=func.now())



from pathlib import Path

import click
from celery import Celery

from celery_heimdall.contrib.inspector.monitor import monitor


@click.group()
def cli():
    """
    heimdall-inspector provides tools for introspecting a live Celery cluster.
    """


@cli.command('monitor')
@click.argument('broker_url')
@click.option(
    '--enable-events',
    default=False,
    is_flag=True,
    help=(
        'Sends a command-and-control message to all Celery workers to start'
        ' emitting worker events before starting the server.'
    )
)
@click.option(
    '--db',
    default='heimdall.db',
    type=click.Path(dir_okay=False, writable=True, path_type=Path),
    help=(
        'Use the provided path to store our sqlite database.'
    )
)
def monitor_command(broker_url: str, enable_events: bool, db: Path):
    """
    Starts a monitor to watch for Celery events and records them to an SQLite
    database.

    Optionally enables event monitoring on a live cluster if --enable-events is
    provided. Note that it will not stop events when finished.
    """
    if enable_events:
        celery_app = Celery(broker=broker_url)
        celery_app.control.enable_events()

    monitor(broker=broker_url, db=db)


if __name__ == '__main__':
    cli()



from celery import Celery
from celery.app import app_or_default


class Config:
    def __init__(self, app: Celery, *, task=None):
        self.app = app_or_default(app)
        self.task = task

    def _from_task_or_app(self, key, default):
        if self.task:
            v = getattr(self.task, 'heimdall', {}).get(key)
            if v is not None:
                return v

        return self.app.conf.get(f'heimdall_{key}', default)

    @property
    def unique_lock_timeout(self):
        return self._from_task_or_app('unique_lock_timeout', 1)

    @property
    def unique_lock_blocking(self):
        return self._from_task_or_app('unique_lock_blocking', True)

    @property
    def unique_timeout(self):
        return self._from_task_or_app('unique_timeout', 60 * 60)

    @property
    def lock_prefix(self):
        return self._from_task_or_app('lock_prefix', 'h-lock:')

    @property
    def rate_limit_prefix(self):
        return self._from_task_or_app('rate_limit_prefix', 'h-rate:')

    @property
    def unique_raises(self):
        return self._from_task_or_app('unique_raises', False)



import dataclasses
import enum
import hashlib
import datetime
import inspect
from abc import ABC
from typing import Union, Tuple, Callable

import redis
import redis.lock
import celery
from kombu import serialization
from kombu.utils import uuid

from celery_heimdall.config import Config
from celery_heimdall.errors import AlreadyQueuedError


class Strategy(enum.Enum):
    DEFAULT = 10


@dataclasses.dataclass
class RateLimit:
    rate_limit: Union[Tuple, Callable]
    strategy: Strategy = Strategy.DEFAULT


def acquire_lock(task: 'HeimdallTask', key: str, timeout: int, *, task_id: str):
    acquired = redis.lock.Lock(
        task.heimdall_redis,
        key,
        timeout=timeout,
        blocking=task.heimdall_config.unique_lock_blocking,
        blocking_timeout=task.heimdall_config.unique_lock_timeout,
    ).acquire(token=task_id)

    if not acquired:
        pipe = task.heimdall_redis.pipeline()
        pipe.get(key)
        pipe.ttl(key)
        task_id, ttl = pipe.execute()

        raise AlreadyQueuedError(
            # TTL may be -1 or -2 if the key didn't exist, depending on the
            # version of Redis.
            expires_in=max(0, ttl),
            likely_culprit=task_id.decode('utf-8') if task_id else None
        )

    return acquired


def release_lock(task: 'HeimdallTask', key: str):
    task.heimdall_redis.delete(key)


def unique_key_for_task(task: 'HeimdallTask', args, kwargs, *,
                        prefix='') -> str:
    """
    Given a task and its arguments, generate a unique key which can be used
    to identify it.
    """
    h = getattr(task, 'heimdall', {})

    # When Celery deserializes the arguments for a job, args and kwargs will
    # be `[]` or `{}`, even if they were `None` when serialized. Ensure we
    # do the same here or the hashes will never match when arguments are empty.
    args = args or []
    kwargs = kwargs or {}

    # User specified an explicit key function.
    if 'key' in h:
        return prefix + h['key'](args, kwargs)

    # Try to generate a unique key from the arguments given to the task.
    # Most of the cases where this will fail are also cases where Celery
    # will be unable to serialize the job, so we're not too concerned with
    # validation.
    _, _, data = serialization.dumps(
        (args, kwargs),
        # TODO: We should _probably_ use the same serializer as the task.
        'json'
    )

    h = hashlib.md5()
    h.update(task.name.encode('utf-8'))
    h.update(data.encode('utf-8'))
    return f'{prefix}{h.hexdigest()}'


def rate_limited_countdown(task: 'HeimdallTask', key, args, kwargs):
    # Based on improvements to Vigrond's original implementation by mlissner
    # on stack overflow.
    h = getattr(task, 'heimdall', {})
    r = task.heimdall_redis

    if 'rate_limit' in h:
        try:
            times, per = h['rate_limit'].rate_limit
        except TypeError as e:
            f = h['rate_limit'].rate_limit

            rate_limit_args = {}
            signature = inspect.signature(f)
            if 'key' in signature.parameters:
                rate_limit_args['key'] = key
            if 'task' in signature.parameters:
                rate_limit_args['task'] = task
            if 'args' in signature.parameters:
                rate_limit_args['args'] = args
            if 'kwargs' in signature.parameters:
                rate_limit_args['kwargs'] = kwargs

            times, per = h['rate_limit'].rate_limit(**rate_limit_args)
    else:
        times, per = h['times'], h['per']

    number_of_running_tasks = r.get(key)
    if number_of_running_tasks is None:
        r.set(key, 1, ex=per)
        return 0

    if int(number_of_running_tasks) < times:
        if r.incr(key, 1) == 1:
            r.expire(key, per)
        return 0

    schedule_key = f'{key}.schedule'
    now = datetime.datetime.now(tz=datetime.timezone.utc)

    delay = r.get(schedule_key)
    if delay is None or int(delay) < now.timestamp():
        # Either not scheduled, or scheduled in the past.
        ttl = r.ttl(key)
        if ttl < 0:
            return 0

        r.set(
            schedule_key,
            int((now + datetime.timedelta(seconds=ttl)).timestamp()),
            ex=ttl + 20
        )
        return ttl

    new_time = (
        datetime.datetime.fromtimestamp(
            int(delay),
            tz=datetime.timezone.utc
        ) + datetime.timedelta(seconds=per // times)
    )
    new_delay = int((new_time - now).total_seconds())
    r.set(schedule_key, int(new_time.timestamp()), ex=new_delay + 20)
    return new_delay


class HeimdallTask(celery.Task, ABC):
    """
    An all-seeing base task for Celery, it provides useful global utilities
    for common Celery behaviors, such as global rate limiting and singleton
    (only one at a time) tasks.
    """
    abstract = True

    def __init__(self):
        super().__init__()
        self._heimdall_config = None
        self._heimdall_redis = None

    @property
    def heimdall_config(self) -> Config:
        if not self._heimdall_config:
            self._heimdall_config = Config(self.app, task=self)
        return self._heimdall_config

    @property
    def heimdall_redis(self) -> redis.Redis:
        if not self._heimdall_redis:
            self._heimdall_redis = self.setup_redis()
        return self._heimdall_redis

    def setup_redis(self) -> redis.Redis:
        """
        Sets up the Redis connection. By default, it'll use any Redis instance
        it can find (in order):

            - the Celery result backend
            - the Celery broker

        If nothing can be found, or if you want to explicitly specify a Redis
        connection you'll need to implement this method yourself, ex:

        .. code::

            from redis import Redis
            from celery_heimdall import HeimdallTask

            class MyHeimdallTask(HeimdallTask):
                def setup_redis(self):
                    return Redis.from_url('redis://')
        """
        # Try to use the Celery result backend, if it's configured for redis.
        backend = self.app.conf.get('result_backend') or ''
        if backend.startswith('redis://'):
            return redis.Redis.from_url(backend)

        # If not the backend, try the broker....
        broker = self.app.conf.get('broker_url') or ''
        if broker.startswith('redis://'):
            return redis.Redis.from_url(broker)

        # Nope, we can't find a usable redis, user will need to implement
        # setup_redis() themselves.
        raise NotImplementedError()

    def apply_async(self, args=None, kwargs=None, task_id=None, **options):
        h = getattr(self, 'heimdall', {})
        if h and 'unique' in h:
            task_id = task_id or uuid()

            # Task has been configured to be globally unique, so we check for
            # the presence of a global lock before allowing it to be queued.
            try:
                acquire_lock(
                    self,
                    unique_key_for_task(
                        self,
                        args,
                        kwargs,
                        prefix=self.heimdall_config.lock_prefix
                    ),
                    h.get(
                        'unique_timeout',
                        self.heimdall_config.unique_timeout
                    ),
                    task_id=task_id
                )
            except AlreadyQueuedError as exc:
                if not self.heimdall_config.unique_raises:
                    # If we were unable to get the task ID for whatever reason,
                    # we just fall through and raise anyway.
                    if exc.likely_culprit is not None:
                        return self.AsyncResult(exc.likely_culprit)

                raise

        # TODO: If we kept track of queued, but not running, tasks, we should
        #       be able to estimate _when_ it would be okay to run a
        #       rate-limited task, rather then just checking when it runs.

        return super().apply_async(
            args=args,
            kwargs=kwargs,
            task_id=task_id,
            **options
        )

    def __call__(self, *args, **kwargs):
        h = getattr(self, 'heimdall', {})
        if h and ('per' in h and 'times' in h) or 'rate_limit' in h:
            delay = rate_limited_countdown(
                self,
                unique_key_for_task(
                    self,
                    args,
                    kwargs,
                    prefix=self.heimdall_config.rate_limit_prefix
                ),
                args,
                kwargs
            )
            if delay > 0:
                # We don't want our rescheduling retry to count against
                # any normal retry limits the user might have set on the
                # task or globally.
                self.request.retries -= 1
                # Max retries needs to be set to None _before_ calling
                # retry(). This value will not propagate, allowing the user's
                # normal retry behaviour to apply on the next call.
                self.max_retries = None
                raise self.retry(countdown=delay)

        return self.run(*args, **kwargs)

    def after_return(self, status, retval, task_id, args, kwargs, einfo):
        # Handles post-task cleanup, when a task exits cleanly. This will be
        # called if a task raises an exception (stored in `einfo`), but not
        # if a worker straight up dies (say, because of running out of memory)
        h = getattr(self, 'heimdall', {})

        # Cleanup the unique task lock when the task finishes, unless the user
        # told us to wait for the remaining interval.
        if h and 'unique' in h and not h.get('unique_wait_for_expiry'):
            release_lock(
                self,
                unique_key_for_task(
                    self,
                    args,
                    kwargs,
                    prefix=self.heimdall_config.lock_prefix
                )
            )

        super().after_return(status, retval, task_id, args, kwargs, einfo)

    def only_after(self, key: str, seconds: int) -> bool:
        """
        A utility for writing sub-blocks in tasks that only execute if
        `seconds` has passed since the last time it was run.

        Imagine you have a task that runs every 5 minutes, but there's one line
        in that task you only want to run after at least an hour. You'd use
        `only_after` to accomplish that.
        """
        task_id = getattr(self.request, 'id', uuid())
        return bool(
            redis.lock.Lock(
                self.heimdall_redis,
                key,
                timeout=seconds,
                blocking=self.heimdall_config.unique_lock_blocking,
                blocking_timeout=self.heimdall_config.unique_lock_timeout
            ).acquire(token=task_id)
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['celery_heimdall',
 'celery_heimdall.contrib',
 'celery_heimdall.contrib.inspector']

package_data = \
{'': ['*']}

install_requires = \
['celery>=5.2.7,<6.0.0', 'importlib-metadata<=4.13', 'redis>=4.3.4,<5.0.0']

extras_require = \
{'inspector': ['click>=8.1.3,<9.0.0', 'SQLAlchemy>=1.4.40,<2.0.0']}

entry_points = \
{'console_scripts': ['heimdall-inspector = '
                     'celery_heimdall.contrib.inspector.cli:cli']}

setup_kwargs = {
    'name': 'celery-heimdall',
    'version': '0.5.0',
    'description': 'Helpful celery extensions.',
    'long_description': '# celery-heimdall\n\n[![codecov](https://codecov.io/gh/TkTech/celery-heimdall/branch/main/graph/badge.svg?token=1A2CVHQ25Q)](https://codecov.io/gh/TkTech/celery-heimdall)\n![GitHub](https://img.shields.io/github/license/tktech/celery-heimdall)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/celery-heimdall)\n\nCelery Heimdall is a set of common utilities useful for the Celery background\nworker framework, built on top of Redis. It\'s not trying to handle every use\ncase, but to be an easy, modern, and maintainable drop-in solution for 90% of\nprojects.\n\n## Features\n\n- Globally unique tasks, allowing only 1 copy of a task to execute at a time, or\n  within a time period (ex: "Don\'t allow queuing until an hour has passed")\n- Global rate limiting. Celery has built-in rate limiting, but it\'s a rate limit\n  _per worker_, making it unsuitable for purposes such as limiting requests to\n  an API.\n\n## Installation\n\n`pip install celery-heimdall`\n\n## Usage\n\n### Unique Tasks\n\nImagine you have a task that starts when a user presses a button. This task\ntakes a long time and a lot of resources to generate a report. You don\'t want\nthe user to press the button 10 times and start 10 tasks. In this case, you\nwant what Heimdall calls a unique task:\n\n```python\nfrom celery import shared_task\nfrom celery_heimdall import HeimdallTask\n\n@shared_task(base=HeimdallTask, heimdall={\'unique\': True})\ndef generate_report(customer_id):\n    pass\n```\n\nAll we\'ve done here is change the base Task class that Celery will use to run\nthe task, and passed in some options for Heimdall to use. This task is now\nunique - for the given arguments, only 1 will ever run at the same time.\n\n#### Expiry\n\nWhat happens if our task dies, or something goes wrong? We might end up in a\nsituation where our lock never gets cleared, called [deadlock][]. To work around\nthis, we add a maximum time before the task is allowed to be queued again:\n\n\n```python\nfrom celery import shared_task\nfrom celery_heimdall import HeimdallTask\n\n@shared_task(\n  base=HeimdallTask,\n  heimdall={\n    \'unique\': True,\n    \'unique_timeout\': 60 * 60\n  }\n)\ndef generate_report(customer_id):\n  pass\n```\n\nNow, `generate_report` will be allowed to run again in an hour even if the\ntask got stuck, the worker ran out of memory, the machine burst into flames,\netc...\n\n#### Custom Keys\n\nBy default, a hash of the task name and its arguments is used as the lock key.\nBut this often might not be what you want. What if you only want one report at\na time, even for different customers? Ex:\n\n```python\nfrom celery import shared_task\nfrom celery_heimdall import HeimdallTask\n\n@shared_task(\n  base=HeimdallTask,\n  heimdall={\n    \'unique\': True,\n    \'key\': lambda args, kwargs: \'generate_report\'\n  }\n)\ndef generate_report(customer_id):\n  pass\n```\nBy specifying our own key function, we can completely customize how we determine\nif a task is unique.\n\n#### The Existing Task\n\nBy default, if you try to queue up a unique task that is already running,\nHeimdall will return the existing task\'s `AsyncResult`. This lets you write\nsimple code that doesn\'t need to care if a task is unique or not. Imagine a\nsimple API endpoint that starts a report when it\'s hit, but we only want it\nto run one at a time. The below is all you need:\n\n```python\nimport time\nfrom celery import shared_task\nfrom celery_heimdall import HeimdallTask\n\n@shared_task(base=HeimdallTask, heimdall={\'unique\': True})\ndef generate_report(customer_id):\n  time.sleep(10)\n\ndef my_api_call(customer_id: int):\n  return {\n    \'status\': \'RUNNING\',\n    \'task_id\': generate_report.delay(customer_id).id\n  }\n```\n\nEverytime `my_api_call` is called with the same `customer_id`, the same\n`task_id` will be returned by `generate_report.delay()` until the original task\nhas completed.\n\nSometimes you\'ll want to catch that the task was already running when you tried\nto queue it again. We can tell Heimdall to raise an exception in this case:\n\n```python\nimport time\nfrom celery import shared_task\nfrom celery_heimdall import HeimdallTask, AlreadyQueuedError\n\n\n@shared_task(\n  base=HeimdallTask,\n  heimdall={\n    \'unique\': True,\n    \'unique_raises\': True\n  }\n)\ndef generate_report(customer_id):\n  time.sleep(10)\n\n\ndef my_api_call(customer_id: int):\n  try:\n    task = generate_report.delay(customer_id)\n    return {\'status\': \'STARTED\', \'task_id\': task.id}\n  except AlreadyQueuedError as exc:\n    return {\'status\': \'ALREADY_RUNNING\', \'task_id\': exc.likely_culprit}\n```\n\nBy setting `unique_raises` to `True` when we define our task, an\n`AlreadyQueuedError` will be raised when you try to queue up a unique task\ntwice. The `AlreadyQueuedError` has two properties:\n\n- `likely_culprit`, which contains the task ID of the already-running task,\n- `expires_in`, which is the time remaining (in seconds) before the \n  already-running task is considered to be expired.\n\n#### Unique Interval Task\n\nWhat if we want the task to only run once in an hour, even if it\'s finished?\nIn those cases, we want it to run, but not clear the lock when it\'s finished:\n\n```python\nfrom celery import shared_task\nfrom celery_heimdall import HeimdallTask\n\n@shared_task(\n  base=HeimdallTask,\n  heimdall={\n    \'unique\': True,\n    \'unique_timeout\': 60 * 60,\n    \'unique_wait_for_expiry\': True\n  }\n)\ndef generate_report(customer_id):\n  pass\n```\n\nBy setting `unique_wait_for_expiry` to `True`, the task will finish, and won\'t\nallow another `generate_report()` to be queued until `unique_timeout` has\npassed.\n\n### Rate Limiting\n\nCelery offers rate limiting out of the box. However, this rate limiting applies\non a per-worker basis. There\'s no reliable way to rate limit a task across all\nyour workers. Heimdall makes this easy:\n\n```python\nfrom celery import shared_task\nfrom celery_heimdall import HeimdallTask, RateLimit\n\n@shared_task(\n  base=HeimdallTask,\n  heimdall={\n    \'rate_limit\': RateLimit((2, 60))\n  }\n)\ndef download_report_from_amazon(customer_id):\n  pass\n```\n\nThis says "every 60 seconds, only allow this task to run 2 times". If a task\ncan\'t be run because it would violate the rate limit, it\'ll be rescheduled.\n\nIt\'s important to note this does not guarantee that your task will run _exactly_\ntwice a second, just that it won\'t run _more_ than twice a second. Tasks are\nrescheduled with a random jitter to prevent the [thundering herd][] problem.\n\n\n#### Dynamic Rate Limiting\n\nJust like you can dynamically provide a key for a task, you can also\ndynamically provide a rate limit based off that key.\n\n\n```python\nfrom celery import shared_task\nfrom celery_heimdall import HeimdallTask, RateLimit\n\n\n@shared_task(\n  base=HeimdallTask,\n  heimdall={\n    # Provide a lower rate limit for the customer with the ID 10, for everyone\n    # else provide a higher rate limit.\n    \'rate_limit\': RateLimit(lambda args: (1, 30) if args[0] == 10 else (2, 30)),\n    \'key\': lambda args, kwargs: f\'customer_{args[0]}\'\n  }\n)\ndef download_report_from_amazon(customer_id):\n  pass\n```\n\n\n## Inspirations\n\nThese are more mature projects which inspired this library, and which may\nsupport older versions of Celery & Python then this project.\n\n- [celery_once][], which is unfortunately abandoned and the reason this project\n  exists.\n- [celery_singleton][]\n- [This snippet][snip] by Vigrond, and subsequent improvements by various\n  contributors.\n\n\n[celery_once]: https://github.com/cameronmaske/celery-once\n[celery_singleton]: https://github.com/steinitzu/celery-singleton\n[deadlock]: https://en.wikipedia.org/wiki/Deadlock\n[thundering herd]: https://en.wikipedia.org/wiki/Thundering_herd_problem\n[snip]: https://gist.github.com/Vigrond/2bbea9be6413415e5479998e79a1b11a',
    'author': 'Tyler Kennedy',
    'author_email': 'tk@tkte.ch',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tktech/celery-heimdall',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import datetime
from pathlib import Path

from celery import Celery
from sqlalchemy import create_engine, insert, func, update
from sqlalchemy.dialects.sqlite import insert

from celery_heimdall.contrib.inspector import models


def task_received(event):
    with models.Session() as session:
        session.execute(
            insert(models.TaskInstance.__table__).values(
                uuid=event['uuid'],
                name=event['name'],
                status=models.TaskStatus.RECEIVED,
                hostname=event['hostname'],
                args=event['args'],
                kwargs=event['kwargs'],
                received=datetime.datetime.fromtimestamp(
                    event['timestamp']
                )
            )
        )
        session.commit()


def task_started(event):
    with models.Session() as session:
        session.execute(
            update(
                models.TaskInstance.__table__
            ).where(
                models.TaskInstance.uuid == event['uuid']
            ).values(
                runtime=event.get('runtime', 0),
                status=models.TaskStatus.STARTED,
                started=datetime.datetime.fromtimestamp(
                    event['timestamp']
                ),
                last_seen=func.now()
            )
        )
        session.commit()


def task_succeeded(event):
    with models.Session() as session:
        session.execute(
            update(
                models.TaskInstance.__table__
            ).where(
                models.TaskInstance.uuid == event['uuid']
            ).values(
                runtime=event.get('runtime', 0),
                status=models.TaskStatus.SUCCEEDED,
                succeeded=datetime.datetime.fromtimestamp(
                    event['timestamp']
                ),
                last_seen=func.now()
            )
        )
        session.commit()


def task_retried(event):
    with models.Session() as session:
        session.execute(
            update(
                models.TaskInstance.__table__
            ).where(
                models.TaskInstance.uuid == event['uuid']
            ).values(
                runtime=event.get('runtime', 0),
                status=models.TaskStatus.RETRIED,
                retries=models.TaskInstance.retries + 1,
                last_seen=func.now()
            )
        )
        session.commit()


def task_failed(event):
    with models.Session() as session:
        session.execute(
            update(
                models.TaskInstance.__table__
            ).where(
                models.TaskInstance.uuid == event['uuid']
            ).values(
                runtime=event.get('runtime', 0),
                status=models.TaskStatus.FAILED,
                failed=datetime.datetime.fromtimestamp(
                    event['timestamp']
                ),
                last_seen=func.now()
            )
        )
        session.commit()


def task_rejected(event):
    with models.Session() as session:
        session.execute(
            update(
                models.TaskInstance.__table__
            ).where(
                models.TaskInstance.uuid == event['uuid']
            ).values(
                runtime=event.get('runtime', 0),
                status=models.TaskStatus.REJECTED,
                rejected=datetime.datetime.fromtimestamp(
                    event['timestamp']
                ),
                last_seen=func.now()
            )
        )
        session.commit()


def worker_event(event):
    field_mapping = {
        'freq': models.Worker.frequency,
        'sw_ident': models.Worker.sw_identity,
        'sw_ver': models.Worker.sw_version,
        'sw_sys': models.Worker.sw_system,
        'active': models.Worker.active,
        'processed': models.Worker.processed
    }

    payload = {
        'last_seen': func.now(),
        'status': {
            'worker-heartbeat': models.WorkerStatus.ALIVE,
            'worker-online': models.WorkerStatus.ALIVE,
            'worker-offline': models.WorkerStatus.OFFLINE,
        }.get(event['type'], models.WorkerStatus.LOST)
    }
    for k, v in field_mapping.items():
        if k in event:
            payload[v] = event[k]

    # FIXME: Support postgres / MySQL
    with models.Session() as session:
        session.execute(
            insert(models.Worker.__table__).values({
                'id': event['hostname'],
                **payload
            }).on_conflict_do_update(
                index_elements=['id'],
                set_=payload
            )
        )
        session.commit()


def monitor(*, broker: str, db: Path):
    """
    A real-time Celery event monitor which captures events and populates a
    supported SQLAlchemy database.
    """
    app = Celery(broker=broker)

    engine = create_engine(f'sqlite:///{db}')
    models.Session.configure(bind=engine)
    models.Base.metadata.create_all(engine)

    with app.connection() as connection:
        recv = app.events.Receiver(
            connection,
            handlers={
                # '*': state.event,
                'task-started': task_started,
                'task-rejected': task_rejected,
                'task-failed': task_failed,
                'task-received': task_received,
                'task-succeeded': task_succeeded,
                'task-retried': task_retried,
                'worker-online': worker_event,
                'worker-heartbeat': worker_event,
                'worker-offline': worker_event
            }
        )
        recv.capture(limit=None, timeout=None, wakeup=True)
